require 'sched'
require 'web.compile'
require 'web.server'
require 'print'
local persist = require 'persist'
local tablecopy = (require'utils.table').copy
local lfs = require'lfs'
local utilsystem = require 'utils.system'
local url = require 'socket.url'
local yajl = require 'yajl'
local posixsignal = require 'sched.posixsignal'

local READ_CHUNK_SIZE   = 128*1024   -- read 128k chunks at a time

local function static_file(url, file)
    local ext = file :match "%.(%w+)$" -- extension (to guess mime type)
    local mime_type = ext and web.mime_types[ext] or web.mime_types["<default>"]
    local function handler(echo, env)
        local file, err = io.open(file, "rb")
        if not file then return web.send_error(env, 500, err) end
        repeat
            local buf, err = file:read(READ_CHUNK_SIZE)
            if not buf and err then return web.send_error(env, 500, err) end
            echo(buf)
        until not buf
        file:close()
        return "ok"
    end
    web.site[url] = {content=handler, mime_type=mime_type}
end

local function static_dir(urlprefix, dir, headers, callback)
    dir = dir or ""
    local function handler(echo, env)
        local i, j = env.url:find(urlprefix)
        local filepath = url.unescape(env.url:sub(j+1):match('/?(.*)'))
        if headers then tablecopy(headers, env.response_headers, true) end
        local file, err = io.open((dir.."/"..filepath), "rb")
        if not file then return web.send_error(env, 500, err) end
        repeat
            local buf, err = file:read(READ_CHUNK_SIZE)
            if not buf and err then return web.send_error(env, 500, err) end
            echo(buf)
        until not buf
        file:close()
        if callback then callback(filepath, env) end
        return "ok"
    end
    web.pattern[urlprefix] = handler -- mime_type derived from the URL
end

local iperf_exec = "iperf"

local config = persist.table.new("config")
local default_config={server="-s", client="-c 127.0.0.1"}


local output = {client={}, server={}}
local running = {client=false, server=false}

local function run_thread(type)
    local cmdline = iperf_exec.." "..config[type]
    table.insert(output[type], "Running: iperf "..config[type])
    running[type] = io.popen(cmdline)
    local f = running[type]
    while true do
        local out = f:read("*l")
        if not out then break end
        table.insert(output[type], out)
    end
    f:close()
    table.insert(output[type], "Process stopped")
    running[type] = false
end

local function start_service(type, env)
    config[type] = env.params.config or default_config[type]
    if not running[type] then sched.run(run_thread, type) end
end
local function stop_service(type)
    if running[type] then
        posixsignal.kill(running[type].getpid()+1, "SIGTERM") -- This is huge hack that needs to fixed. popen launches the command with "sh -c"
                                                              -- so the child process has another pid, which is most of the time +1 because it is
                                                              -- launched just after sh... A possible fix is to implement another popen that only execute
                                                              -- with the given name (i.e. without the indirection)
        local count = 0
        while running[type] do
            sched.wait(0.2)
            count = count + 1
            if count > 5 then
                print("Could not terminate properly, try a SIGKILL signal")
                posixsignal.kill(running[type].getpid()+1, "SIGKILL") break
            end
        end
    end
end

local function status(type, echo, env)
    local c = env.params.first_update=="true" and 1 or output[type].n or 1
    local n = #output[type]
    local out
    if c <= n then out = table.concat(output[type], "\r\n", c, n).."\r\n" end
    output[type].n = n+1
    local data = {out = out, running = (running[type] and true or false), config = config[type] or default_config[type]}
    echo(yajl.to_string(data))
end

local function flush_output(type)
    output[type] = {}
end


function main()

    -- Determine the hardware platform in order to execute the correct iperf executable
    local env = tonumber(os.getenv("DEVICE_FAMILY"))
    if env then
        if env <= 3 then iperf_exec="./iperf-pxa"
        else iperf_exec="./iperf-imx" end
    end

    web.mime_types["css"] = "text/css"
    web.mime_types["json"] = "application/json"

    static_dir("static/", "static")
    static_file("", "static/index.html")
    static_file("manual", "static/manual.html")

    static_file("server", "static/server.html")
    web.site["server/start"] =   web.redirect("/server", function(...) return start_service("server", ...) end)
    web.site["server/stop"] =  function(...) return stop_service("server", ...) end
    web.site["server/flush"] =  function() return flush_output("server") end
    web.site["server/status.json"] = function(...) return status("server", ...) end

    static_file("client", "static/client.html")
    web.site["client/start"] =   web.redirect("/client", function(...) return start_service("client", ...) end)
    web.site["client/stop"] =  function(...) return stop_service("client", ...) end
    web.site["client/flush"] =  function() return flush_output("client") end
    web.site["client/status.json"] = function(...) return status("client", ...) end

    web.start("*", 8081)
end

sched.run(main)
sched.loop()




